==> pan-0.6.5.tar.bz2: pan-0.6.5/pan/pan-db.c <==
/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 TODO:
 [ ] transparent (de)compression of database (bzip2 preferably)
*/

#include <errno.h>
#include "../config.h"
#include "pan.h"

/* needed for db.h with 'gcc -ansi -pedantic' */
#ifndef _BSD_SOURCE
#  define _BSD_SOURCE 1
#endif
#include <sys/types.h>
#include <sys/fcntl.h>

#ifdef PREFER_DB1
#ifdef HAVE_DB1_DB_H
# include <db1/db.h>
#else
# ifdef HAVE_DB_185_H
#  include <db_185.h>
# else
#  include <db.h>
# endif
#endif
#else
#ifdef HAVE_DB_185_H
# include <db_185.h>
#else
# ifdef HAVE_DB_H
#  include <db.h>
# else
#  include <db1/db.h>
# endif
#endif
#endif


static
gboolean
open_database (
	const gchar* filename,
	u_int32_t flags,
	DB** setme )
{
	*setme = dbopen ( filename, flags|O_CREAT|O_RDWR, 0700, DB_BTREE, NULL );
	if ( !*setme )
		g_warning ( "Can't open database ``%s'': %s", filename, strerror(errno) );
	return ( *setme != NULL );
}


static void
clear_dbt (
	DBT* setme )
{
	g_assert ( setme!=0 );
	memset ( setme, 0, sizeof(DBT) );
}

static void
set_dbt (
	const char* str,
	DBT* setme )
{
	g_assert ( str!=0 && *str!=0 );
	g_assert ( setme!=0 );

	clear_dbt ( setme );
	setme->data = (char*)str;
	setme->size = strlen(str) + 1;
}
	

//============================================================================

typedef struct
{
	DB* db;
	gchar* filename;
	int ref_count;
}
PanDB;

pan_db
pan_db_open (
	const char* filename )
{
	PanDB* pdb = g_malloc0 ( sizeof ( PanDB ) );
	pdb->filename = g_strdup ( filename );
	pdb->ref_count = 0;
	open_database ( filename, 0, &pdb->db );
	return (pan_db)pdb;
}

void
pan_db_close (
	pan_db db )
{
	PanDB* pdb = (PanDB*)db;
	g_assert ( !pdb->ref_count );
	pdb->db->close ( pdb->db );
	g_free ( pdb->filename );
	g_free ( pdb );
}

gchar*
pan_db_get_value_str (
	pan_db db,
	const char* key_str )
{
	DB* real_db = ((PanDB*)db)->db;
	gchar* retval = NULL;
	DBT key, val;
	set_dbt ( key_str, &key );
	clear_dbt ( &val );
	if ( !db )
		return NULL;
	if ( !real_db->get ( real_db, &key, &val, 0 ) )
		retval = g_strdup ( (char*)(val.data) );
	return retval;
}

void
pan_db_put_value_str (
	pan_db db,
	const char* key_str,
	const char* val_str )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key, val;
	set_dbt ( key_str, &key );
	set_dbt ( val_str, &val );
	real_db->put ( real_db, &key, &val, 0 );
}

int
pan_db_get_value_i (
	pan_db db,
	const char* key )
{
	gchar* val = pan_db_get_value_str ( db, key );
	int retval = val ? atoi ( val ) : 0;
	g_free ( val );
	return retval;
}

void
pan_db_put_value_i (
	pan_db db,
	const char* key,
	int i )
{
	char buf [ 32 ];
	sprintf ( buf, "%d", i );
	pan_db_put_value_str ( db, key, buf );
}

void
pan_db_add_value_i (
	pan_db db,
	const char* key,
	int i )
{
	pan_db_put_value_i ( db, key, i + pan_db_get_value_i ( db, key ) );
}

void
pan_db_erase (
	pan_db db,
	const char* key_str )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key;
	int status;
	set_dbt ( key_str, &key );
        status = real_db->del ( real_db, &key, 0 );
	if ( status )
		g_warning ( "Unable to erase record %s: %s", key_str, strerror(status) );
}

void
pan_db_foreach (
	pan_db db,
	DBFunc func,
	gpointer user_data )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key, val;
	int status;

	clear_dbt ( &key );
	clear_dbt ( &val );
	while ( ( status = real_db->seq ( real_db, &key, &val, R_NEXT ) ) == 0 )
		(*func)(key.data, val.data, user_data );
}

void
pan_db_sync (
	pan_db db )
{
	DB* real_db = ((PanDB*)db)->db;
	real_db->sync ( real_db, 0 );
}

void
pan_db_erase_all (
	pan_db db )
{
	PanDB* pdb = (PanDB*)db;
	pdb->db->close ( pdb->db );
        remove ( pdb->filename );
	open_database ( pdb->filename, 0, &pdb->db );
}

//============================================================================

static GHashTable* db_hash = NULL;

pan_db
pan_db_ref (
	const char* filename )
{
	pan_db db = NULL;

	/* init the hash table if necessary */
	if ( !db_hash )
		db_hash = g_hash_table_new ( g_str_hash, g_str_equal );

	/* get the db, or create a new one if we don't have one already */
	db = g_hash_table_lookup ( db_hash, filename );
	if ( !db ) {
		db = pan_db_open ( filename );
		g_hash_table_insert ( db_hash, g_strdup(filename), db ); /* FIXME: key leak */
	}

	/* pop up the ref count and go */
	++((PanDB*)db)->ref_count;
	return db;
}

void
pan_db_unref (
	pan_db db )
{
	PanDB* pan_db = (PanDB*)db;

	g_return_if_fail ( pan_db->ref_count >= 0 );
	g_return_if_fail ( db!=NULL ); /* don't pass in null */
	g_return_if_fail ( db_hash!=NULL ); /* created in ref */
	g_return_if_fail ( g_hash_table_lookup ( db_hash, pan_db->filename ) == db );

	if ( !--pan_db->ref_count ) {
		g_hash_table_remove ( db_hash, pan_db->filename );
		pan_db_close ( db );
	}
}

==> pan-0.6.6.tar.bz2: pan-0.6.6/pan/pan-db.c <==
/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 * pan-db (C) 1999 Charles Kerr (charles@skywalker.ecn.ou.edu)
 *
 * Author: Charles Kerr (charles@skywalker.ecn.ou.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 TODO:
 [ ] transparent (de)compression of database (bzip2 preferably)
*/

#include <errno.h>
#include "../config.h"
#include "pan.h"

/* needed for db.h with 'gcc -ansi -pedantic' */
#ifndef _BSD_SOURCE
#  define _BSD_SOURCE 1
#endif
#include <sys/types.h>
#include <sys/fcntl.h>

#ifdef PREFER_DB1
#ifdef HAVE_DB1_DB_H
# include <db1/db.h>
#else
# ifdef HAVE_DB_185_H
#  include <db_185.h>
# else
#  include <db.h>
# endif
#endif
#else
#ifdef HAVE_DB_185_H
# include <db_185.h>
#else
# ifdef HAVE_DB_H
#  include <db.h>
# else
#  include <db1/db.h>
# endif
#endif
#endif


static
gboolean
open_database (
	const gchar* filename,
	u_int32_t flags,
	DB** setme )
{
	*setme = dbopen ( filename, flags|O_CREAT|O_RDWR, 0700, DB_BTREE, NULL );
	if ( !*setme )
		g_warning ( "Can't open database ``%s'': %s", filename, strerror(errno) );
	return ( *setme != NULL );
}


static void
clear_dbt (
	DBT* setme )
{
	g_assert ( setme!=0 );
	memset ( setme, 0, sizeof(DBT) );
}

static void
set_dbt (
	const char* str,
	DBT* setme )
{
	g_assert ( str!=0 && *str!=0 );
	g_assert ( setme!=0 );

	clear_dbt ( setme );
	setme->data = (char*)str;
	setme->size = strlen(str) + 1;
}
	

//============================================================================

typedef struct
{
	DB* db;
	gchar* filename;
	int ref_count;
}
PanDB;

pan_db
pan_db_open (
	const char* filename )
{
	PanDB* pdb = g_malloc0 ( sizeof ( PanDB ) );
	pdb->filename = g_strdup ( filename );
	pdb->ref_count = 0;
	open_database ( filename, 0, &pdb->db );
	return (pan_db)pdb;
}

void
pan_db_close (
	pan_db db )
{
	PanDB* pdb = (PanDB*)db;
	g_assert ( !pdb->ref_count );
	pdb->db->close ( pdb->db );
	g_free ( pdb->filename );
	g_free ( pdb );
}

gchar*
pan_db_get_value_str (
	pan_db db,
	const char* key_str )
{
	DB* real_db = ((PanDB*)db)->db;
	gchar* retval = NULL;
	DBT key, val;
	set_dbt ( key_str, &key );
	clear_dbt ( &val );
	if ( !db )
		return NULL;
	if ( !real_db->get ( real_db, &key, &val, 0 ) )
		retval = g_strdup ( (char*)(val.data) );
	return retval;
}

void
pan_db_put_value_str (
	pan_db db,
	const char* key_str,
	const char* val_str )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key, val;
	set_dbt ( key_str, &key );
	set_dbt ( val_str, &val );
	real_db->put ( real_db, &key, &val, 0 );
}

int
pan_db_get_value_i (
	pan_db db,
	const char* key )
{
	gchar* val = pan_db_get_value_str ( db, key );
	int retval = val ? atoi ( val ) : 0;
	g_free ( val );
	return retval;
}

void
pan_db_put_value_i (
	pan_db db,
	const char* key,
	int i )
{
	char buf [ 32 ];
	sprintf ( buf, "%d", i );
	pan_db_put_value_str ( db, key, buf );
}

void
pan_db_add_value_i (
	pan_db db,
	const char* key,
	int i )
{
	pan_db_put_value_i ( db, key, i + pan_db_get_value_i ( db, key ) );
}

void
pan_db_erase (
	pan_db db,
	const char* key_str )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key;
	int status;
	set_dbt ( key_str, &key );
        status = real_db->del ( real_db, &key, 0 );
	if ( status )
		g_warning ( "Unable to erase record %s: %s", key_str, strerror(status) );
}

void
pan_db_foreach (
	pan_db db,
	DBFunc func,
	gpointer user_data )
{
	DB* real_db = ((PanDB*)db)->db;
	DBT key, val;
	int status;
	u_int step = R_FIRST;

	clear_dbt ( &key );
	clear_dbt ( &val );
	while ( ( status = real_db->seq ( real_db, &key, &val, step ) ) == 0 ) {
		(*func)(key.data, val.data, user_data );
		step = R_NEXT;
	}
}

void
pan_db_sync (
	pan_db db )
{
	DB* real_db = ((PanDB*)db)->db;
	real_db->sync ( real_db, 0 );
}

void
pan_db_erase_all (
	pan_db db )
{
	PanDB* pdb = (PanDB*)db;
	pdb->db->close ( pdb->db );
        remove ( pdb->filename );
	open_database ( pdb->filename, 0, &pdb->db );
}

//============================================================================

static GHashTable* db_hash = NULL;

pan_db
pan_db_ref (
	const char* filename )
{
	pan_db db = NULL;

	/* init the hash table if necessary */
	if ( !db_hash )
		db_hash = g_hash_table_new ( g_str_hash, g_str_equal );

	/* get the db, or create a new one if we don't have one already */
	db = g_hash_table_lookup ( db_hash, filename );
	if ( !db ) {
		db = pan_db_open ( filename );
		g_hash_table_insert ( db_hash, g_strdup(filename), db ); /* FIXME: key leak */
	}

	/* pop up the ref count and go */
	++((PanDB*)db)->ref_count;
	return db;
}

void
pan_db_unref (
	pan_db db )
{
	PanDB* pan_db = (PanDB*)db;

	g_return_if_fail ( pan_db->ref_count >= 0 );
	g_return_if_fail ( db!=NULL ); /* don't pass in null */
	g_return_if_fail ( db_hash!=NULL ); /* created in ref */
	g_return_if_fail ( g_hash_table_lookup ( db_hash, pan_db->filename ) == db );

	if ( !--pan_db->ref_count ) {
		g_hash_table_remove ( db_hash, pan_db->filename );
		pan_db_close ( db );
	}
}
